"""The model Observer model."""

import uuid
from time import sleep, time
from typing import Any, Iterable, List, Optional, Set, Tuple

from asgiref.sync import async_to_sync
from channels.layers import get_channel_layer
from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.db import models, transaction
from django.db.models import Count, Q
from django.db.models.query import QuerySet

from resolwe.flow.models.base import BaseManagerWithoutVersion
from resolwe.permissions.models import Permission, PermissionObject

from .protocol import GROUP_SESSIONS, TYPE_ITEM_UPDATE, ChangeType, ChannelsMessage

# Type alias for observable object.
Observable = PermissionObject


def _default_output_value():
    """Return the default value for output field in the BackgroundTask model.

    Must be named method due to Django bug in migrations.
    """
    return ""


class BackgroundTask(Observable):
    """The observable model representing the background task.

    When a long running task is started, the instance of type BackgroundTask is created
    and frontend can subscribe to receive its notifications.
    """

    #: background task is waiting
    STATUS_WAITING = "WT"
    #: background task is processing
    STATUS_PROCESSING = "PR"
    #: background task has finished successfully
    STATUS_DONE = "OK"
    #: background task has finished with error
    STATUS_ERROR = "ER"

    STATUS_CHOICES = (
        (STATUS_WAITING, "Waiting"),
        (STATUS_PROCESSING, "Processing"),
        (STATUS_DONE, "Done"),
        (STATUS_ERROR, "Error"),
    )

    objects = BaseManagerWithoutVersion()

    #: task start date and time
    started = models.DateTimeField(blank=True, null=True, db_index=True)

    #: task finished date and time
    finished = models.DateTimeField(blank=True, null=True, db_index=True)

    #: status of the background task
    status = models.CharField(
        max_length=2, choices=STATUS_CHOICES, default=STATUS_WAITING
    )

    #: task description
    description = models.CharField(max_length=256)

    #: the field containing output from the backgroud task, such as list of ids of
    #: duplicated objects, error details...
    output = models.JSONField(default=_default_output_value)

    def wait(
        self,
        timeout: float = 2,
        final_statuses: list[str] = [STATUS_DONE],
        polling_interval: float = 0.3,
    ):
        """Wait for up to timeout seconds for task to transition into final_statuses.

        The method is meant to be used in tests as it creates lots of database queries.

        :raises RuntimeError: when desired status was not reached within timeout.
        """
        started = time()
        while self.status not in final_statuses and time() - started < timeout:
            sleep(polling_interval)
            self.refresh_from_db()

        if self.status not in final_statuses:
            raise RuntimeError(
                f"Task failed to reach {final_statuses} within {timeout} seconds."
            )

    def result(
        self,
        timeout: float = 2,
        final_statuses: list[str] = [STATUS_DONE],
        polling_interval: float = 0.3,
    ) -> Any:
        """Return the output field of the background task when finished.

        The method waits for up to timeout seconds for task to reach the one of
        final_statuses and returns its result (output field).

        :raises RuntimeError: when desired status was not reached within the timeout.
        """
        self.wait(timeout, final_statuses, polling_interval)
        return self.output


def get_random_uuid() -> str:
    """Generate a random UUID in string format."""
    return uuid.uuid4().hex


class Observer(models.Model):
    """An observer for detecting changes in a model.

    Characterized by a db table or an instance of a model and a change type.
    Several subscriptions can subscribe to the same observer.
    """

    ALL_IDS = 0  # The constant used for catchall object id.
    CHANGE_TYPES = [(change.value, change.name) for change in ChangeType]

    #: table of the observed resource
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    #: primary key of the observed resource (null if watching the whole table)
    object_id = models.PositiveIntegerField(null=False)
    #: the type of change to observe for
    change_type = models.PositiveSmallIntegerField(choices=CHANGE_TYPES)

    class Meta:
        """Add index to session_id field and set defining fields."""

        indexes = [models.Index(fields=["object_id"])]
        unique_together = ("content_type", "object_id", "change_type")

    @classmethod
    def get_interested(
        cls,
        content_type: ContentType,
        object_id: int,
        change_type: Optional[ChangeType] = None,
    ) -> "QuerySet[Observer]":
        """Find all observers watching for changes of a given item/table."""
        query = Q(content_type=content_type)
        if change_type is not None:
            query &= Q(change_type=change_type.value)
        query &= Q(object_id=object_id) | Q(object_id=Observer.ALL_IDS)
        return cls.objects.filter(query)

    @classmethod
    def observe_instance_container(
        cls,
        instance: Observable,
        change_type: ChangeType,
        containers: Optional[Iterable[Observable]] = None,
    ):
        """Handle a notifications to the containers of the given instance.

        When containers are given notify them else infer them from the instance.
        """
        # Test explicitely for None, since containers may be empty.
        if containers is None:
            containers = instance.containers
        for container in containers:
            observers = Observer.get_interested(
                change_type=change_type,
                content_type=ContentType.objects.get_for_model(container),
                object_id=container.pk,
            )
            # Forward the message to the appropriate groups.
            for subscriber in Subscription.objects.filter(observers__in=observers):
                if container.has_permission(Permission.VIEW, subscriber.user):
                    # Register on_commit callbacks to send the signals.
                    Subscription.notify(
                        subscriber.session_id,
                        container,
                        change_type,
                        source=(
                            ContentType.objects.get_for_model(instance).name,
                            instance.pk,
                        ),
                    )

    @classmethod
    def observe_instance_changes(cls, instance: Observable, change_type: ChangeType):
        """Handle a notification about an instance change."""
        content_type = ContentType.objects.get_for_model(instance)
        observers = Observer.get_interested(
            change_type=change_type,
            content_type=content_type,
            object_id=instance.pk,
        )

        # Forward the message to the appropriate groups.
        for subscriber in Subscription.objects.filter(observers__in=observers):
            if instance.has_permission(Permission.VIEW, subscriber.user):
                # Register on_commit callbacks to send the signals.
                Subscription.notify(
                    subscriber.session_id,
                    instance,
                    change_type,
                    source=(content_type.name, instance.pk),
                )

    @classmethod
    def observe_permission_changes(
        cls,
        instance: Any,
        gains: Set[int],
        losses: Set[int],
        observe_containers: bool = True,
    ):
        """Handle a notification about a permission change.

        Given an instance and a set of user_ids who gained/lost permissions for it,
        only relevant observers will be notified of the instance's creation/deletion.

        When observe_containers is set to True (default), also object's containers are
        notified using the same change type.
        """
        for change_type, user_ids in (
            (ChangeType.CREATE, gains),
            (ChangeType.DELETE, losses),
        ):
            # A shortcut if nothing actually changed.
            if len(user_ids) == 0:
                continue

            # Find all sessions who have observers registered on this object.
            content_type = ContentType.objects.get_for_model(instance)
            interested = Observer.get_interested(
                change_type=change_type,
                content_type=content_type,
                object_id=instance.pk,
            )
            # Of all interested users, select only those whose permissions changed.
            session_ids = set(
                Subscription.objects.filter(observers__in=interested)
                .filter(user__in=user_ids)
                .values_list("session_id", flat=True)
                .distinct()
            )

            for session_id in session_ids:
                Subscription.notify(
                    session_id,
                    instance,
                    change_type,
                    source=(content_type.name, instance.pk),
                )

            if observe_containers:
                cls.observe_instance_container(instance, change_type)

    def __str__(self) -> str:
        """Format the object representation."""
        return f"content_type={self.content_type} object_id={self.object_id} change={self.change_type}"


class Subscription(models.Model):
    """Subscription to several observers.

    One subscription corresponds to a single api call to the subscribe endpoint.
    A client may have several subscriptions with different ID's to differentiate
    between them.
    """

    #: observers to whom the subscription is listening
    observers = models.ManyToManyField("Observer", related_name="subscriptions")
    #: subscriber's user reference
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
    )
    #: subscription time
    created = models.DateTimeField(auto_now_add=True)

    #: ID of the websocket session (one session may have multiple observers)
    session_id = models.CharField(max_length=100)
    #: unique ID for the client to remember which subscription a signal belongs to
    subscription_id = models.UUIDField(
        unique=True, default=get_random_uuid, editable=False
    )

    class Meta:
        """Add index to session_id field."""

        indexes = [models.Index(fields=["session_id"])]

    def subscribe(
        self,
        content_type: ContentType,
        object_ids: List[int],
        change_types: List[ChangeType],
    ):
        """Assign self to multiple observers at once."""
        for id in object_ids:
            for change_type in change_types:
                observer, _ = Observer.objects.get_or_create(
                    content_type=content_type,
                    object_id=id,
                    change_type=change_type.value,
                )
                self.observers.add(observer)

    @transaction.atomic
    def delete(self):
        """Delete the given subscription.

        Delete all observers with no remaining subscriptions.

        This method is marked atomic, so no new subsciptions are added to the observers
        we are deleting in the meantime causing IntegrityError.
        """
        # First find observers with only this subscription and delete them.
        Observer.objects.annotate(subs=Count("subscriptions")).filter(
            subscriptions=self.pk, subs=1
        ).delete()
        # Now delete the subscription itself.
        super().delete()

    @classmethod
    def notify(
        cls,
        session_id: str,
        instance: Observable,
        change_type: ChangeType,
        source: Optional[Tuple[str, int]],
    ):
        """Register a callback to send a change notification on transaction commit."""
        notification: ChannelsMessage = {
            "type": TYPE_ITEM_UPDATE,
            "content_type_pk": ContentType.objects.get_for_model(instance).pk,
            "change_type_value": change_type.value,
            "object_id": instance.pk,
            "source": source,
        }

        # Define a callback, but copy variable values.
        def trigger(
            channel_layer=get_channel_layer(),
            channel=GROUP_SESSIONS.format(session_id=session_id),
            notification=notification,
        ):
            async_to_sync(channel_layer.group_send)(channel, notification)

        transaction.on_commit(trigger)

    def notify_created(self, content_type: ContentType):
        """Send a create notification.

        Used to send signal when models without permissions are created.
        """
        notification: ChannelsMessage = {
            "type": TYPE_ITEM_UPDATE,
            "content_type_pk": content_type.pk,
            "change_type_value": ChangeType.CREATE.value,
            "object_id": Observer.ALL_IDS,
            "source": None,
        }
        channel_layer = get_channel_layer()
        channel = GROUP_SESSIONS.format(session_id=self.session_id)
        async_to_sync(channel_layer.group_send)(channel, notification)



# Generated by Django 5.1.5 on 2025-02-12 06:15

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ("flow", "0035_move_version_field_to_resolwe"),
    ]

    operations = [
        migrations.AddField(
            model_name="annotationvalue",
            name="deleted",
            field=models.BooleanField(default=False),
        ),
    ]



"""Custom permissions for Flow API."""

from contextlib import suppress
from logging import getLogger
from typing import Optional

from django.contrib.auth.models import User
from django.db import models
from django.http import Http404, HttpRequest
from rest_framework import exceptions, permissions

from resolwe.permissions.models import Permission, PermissionList, get_anonymous_user
from resolwe.permissions.utils import get_user, model_has_permissions

logger = getLogger(__name__)


class ResolwePermissionBackend:
    """Custom resolwe permission backend."""

    supports_object_permissions = True
    supports_anonymous_user = True
    supports_inactive_user = True

    def authenticate(
        self,
        request: HttpRequest,
        username: Optional[User] = None,
        password: Optional[str] = None,
    ):
        """Authenticate user.

        Since our backend is only used for authorization return None. See

        https://docs.djangoproject.com/en/3.2/topics/auth/customizing/

        for details.
        """
        return None

    def has_perm(
        self,
        user: User,
        permission: Permission | str,
        obj: Optional[models.Model] = None,
    ) -> bool:
        """
        Check if user has the given permission on the given object.

        :returns: ``True`` if given ``user_obj`` has ``perm`` for ``obj``.
            If no ``obj`` is given or user is inactive and authenticated at the
            same time ``False`` is returned.
        """
        user = get_user(user)
        if not user.is_active:
            return False

        # We do not handle model level permissions.
        if obj is None:
            return False

        assert isinstance(permission, Permission)
        if not model_has_permissions(obj):
            return True
        else:
            return obj.has_permission(permission, user)

    def get_all_permissions(
        self, user: User, obj: Optional[models.Model] = None
    ) -> PermissionList:
        """Return a set of permissions for ``user`` on ``obj``.

        Permissions from user groups are also considered.

        When user is not active the empty set of permissions is returned.
        """
        all_permissions = []
        if user.is_active and model_has_permissions(obj):
            permission_value = obj.permission_group.permissions.filter(
                models.Q(user=user) | models.Q(group__in=user.groups.all())
            ).aggregate(permission=models.Max("value"))["permission"]
            if permission_value:
                all_permissions = list(Permission(permission_value))
        return all_permissions


class ResolwePermissions(permissions.DjangoObjectPermissions):
    """Resolwe Permissions."""

    perms_map = {
        "GET": [Permission.VIEW],
        "OPTIONS": [Permission.VIEW],
        "HEAD": [Permission.VIEW],
        "POST": [Permission.EDIT],
        "PUT": [Permission.EDIT],
        "PATCH": [Permission.EDIT],
        "DELETE": [Permission.EDIT],
    }

    def get_required_object_permissions(self, method, model_cls):
        """Return the list of permissions that the user is required to have."""
        if method not in self.perms_map:
            raise exceptions.MethodNotAllowed(method)

        return self.perms_map[method]

    def has_permission(self, request, view):
        """Return `True` as we don't use model level perms."""
        return True

    def has_object_permission(self, request, view, obj):
        """Check object permissions."""

        # admins can do anything
        if request.user.is_superuser:
            return True

        # No permission on objects that do not support it.
        if not model_has_permissions(obj):
            with suppress(Exception):
                logger.debug(
                    "Object %s of type %s does not support permissions, 403 (user %s).",
                    obj.pk,
                    obj._meta.model,
                    request.user.pk,
                )

            return False

        # `share` permission is required for editing permissions
        if view.action is not None and "permissions" in view.action:
            self.perms_map["POST"] = [Permission.SHARE]

        if hasattr(view, "get_queryset"):
            queryset = view.get_queryset()
        else:
            queryset = getattr(view, "queryset", None)

        assert queryset is not None, (
            "Cannot apply DjangoObjectPermissions on a view that "
            "does not set `.queryset` or have a `.get_queryset()` method."
        )

        model_cls = queryset.model
        user = request.user
        perm = max(self.get_required_object_permissions(request.method, model_cls))
        logger.debug(
            "Checking perm %s, method %s, user %s, object %s, type %s.",
            perm,
            request.method,
            user.pk,
            obj.pk,
            model_cls,
        )
        anonymous = get_anonymous_user()
        if not user.has_perm(perm, obj) and not anonymous.has_perm(perm, obj):
            if request.method not in permissions.SAFE_METHODS:
                # If the user does not have permissions we need to determine if
                # they have read permissions to see 403, or not, and simply see
                # a 404 response.
                read_perm = max(self.get_required_object_permissions("GET", model_cls))
                logger.debug("Failed, required read permission: %s.", read_perm)

                if user.has_perm(read_perm, obj):
                    logger.debug(
                        "User %s has read permission for object %s of type %s, returning 403.",
                        user.pk,
                        obj.pk,
                        model_cls,
                    )
                    return False

            raise Http404

        return True



"""Entity viewset."""

from logging import getLogger
from typing import Optional

from django.core.exceptions import ValidationError
from django.db.models import F, Func, OuterRef, Prefetch, Subquery
from django.db.models.functions import Coalesce
from django.http import Http404
from django.shortcuts import get_object_or_404 as _get_object_or_404
from drf_spectacular.utils import extend_schema
from rest_framework import exceptions, serializers, status
from rest_framework.decorators import action
from rest_framework.request import Request
from rest_framework.response import Response

from resolwe.flow.filters import EntityFilter
from resolwe.flow.models import Data, Entity
from resolwe.flow.serializers import EntitySerializer
from resolwe.flow.serializers.annotations import AnnotationsByPathSerializer
from resolwe.observers.mixins import ObservableMixin
from resolwe.observers.views import BackgroundTaskSerializer

from .collection import BaseCollectionViewSet
from .utils import get_collection_for_user

logger = getLogger(__name__)


class MoveEntityToCollectionSerializer(serializers.Serializer):
    """Serializer for moving entities to a collection."""

    ids = serializers.ListField(child=serializers.IntegerField())
    destination_collection = serializers.IntegerField()


class EntityViewSet(ObservableMixin, BaseCollectionViewSet):
    """API view for entities."""

    serializer_class = EntitySerializer
    filterset_class = EntityFilter

    data_count_subquery = (
        Data.objects.filter(entity_id=OuterRef("pk"))
        .annotate(count=Func(F("id"), function="Count"))
        .values("count")
    )
    data_status_subquery = (
        Data.objects.filter(entity_id=OuterRef("pk"))
        .annotate(
            count=Coalesce(
                Func(
                    F("status"),
                    function="ARRAY_AGG",
                    template="%(function)s(DISTINCT %(expressions)s)",
                ),
                [],
            )
        )
        .values("count")
    )
    queryset = (
        Entity.objects.select_related("contributor")
        .prefetch_related(
            Prefetch("collection", queryset=BaseCollectionViewSet.queryset)
        )
        .annotate(data_statuses=Subquery(data_status_subquery))
        .annotate(data_count=Subquery(data_count_subquery))
        .annotate_status()
    )

    def _get_entities(self, user, ids):
        """Return entities queryset based on provided entity ids."""
        queryset = Entity.objects.filter(id__in=ids).filter_for_user(user)
        actual_ids = queryset.values_list("id", flat=True)
        missing_ids = list(set(ids) - set(actual_ids))
        if missing_ids:
            raise exceptions.ParseError(
                "Entities with the following ids not found: {}".format(
                    ", ".join(map(str, missing_ids))
                )
            )

        return queryset

    @extend_schema(
        request=MoveEntityToCollectionSerializer(), responses={status.HTTP_200_OK: None}
    )
    @action(detail=False, methods=["post"])
    def move_to_collection(self, request, *args, **kwargs):
        """Move samples from source to destination collection."""

        serializer = MoveEntityToCollectionSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        ids = serializer.validated_data["ids"]
        dst_collection_id = serializer.validated_data["destination_collection"]
        dst_collection = get_collection_for_user(dst_collection_id, request.user)
        entity_qs = self._get_entities(request.user, ids)
        task = entity_qs.move_to_collection(dst_collection, request.user)
        return Response(
            status=status.HTTP_200_OK,
            data=BackgroundTaskSerializer(task).data,
        )

    # NOTE: This can be deleted when DRF will support select_for_update
    #       on updates and ResolweUpdateModelMixin will use it.
    #       https://github.com/encode/django-rest-framework/issues/4675
    def update(self, request, *args, **kwargs):
        """Update an entity.

        Original queryset produces a temporary database table whose rows
        cannot be selected for an update. As a workaround, we patch
        get_queryset function to return only Entity objects without
        additional data that is not needed for the update.
        """
        orig_get_queryset = self.get_queryset

        def patched_get_queryset():
            """Patched get_queryset method."""
            entity_ids = orig_get_queryset().values_list("id", flat=True)
            return Entity.objects.filter(id__in=entity_ids)

        self.get_queryset = patched_get_queryset

        resp = super().update(request, *args, **kwargs)

        self.get_queryset = orig_get_queryset
        return resp

    def get_object(self):
        """Return the object the view is displaying.

        You may want to override this if you need to provide non-standard
        queryset lookups.  Eg if objects are referenced using multiple
        keyword arguments in the url conf.
        """

        def get_object_or_404(queryset, *filter_args, **filter_kwargs):
            """Implement custom get_object_or_404.

            Same as Django's standard shortcut, but make sure to also raise 404
            if the filter_kwargs don't match the required types.
            """
            try:
                return _get_object_or_404(queryset, *filter_args, **filter_kwargs)
            except (TypeError, ValueError, ValidationError):
                raise Http404

        queryset = self.filter_queryset(self.get_queryset())
        logger.debug("Get entity: queryset after filter")

        # Perform the lookup filtering.
        lookup_url_kwarg = self.lookup_url_kwarg or self.lookup_field

        assert lookup_url_kwarg in self.kwargs, (
            "Expected view %s to be called with a URL keyword argument "
            'named "%s". Fix your URL conf, or set the `.lookup_field` '
            "attribute on the view correctly."
            % (self.__class__.__name__, lookup_url_kwarg)
        )

        filter_kwargs = {self.lookup_field: self.kwargs[lookup_url_kwarg]}
        obj = get_object_or_404(queryset, **filter_kwargs)
        logger.debug("Get entity: after 404")

        # May raise a permission denied
        self.check_object_permissions(self.request, obj)

        return obj

    @extend_schema(
        request=AnnotationsByPathSerializer(many=True),
        responses={status.HTTP_200_OK: None},
    )
    @action(detail=True, methods=["post"])
    def set_annotations(self, request: Request, pk: Optional[int] = None):
        """Add the given list of AnnotationFields to the given entity."""
        # No need to check for permissions, since post requires edit by default.
        try:
            logger.debug(
                "User %s called set_annotations for entity %s.", request.user.pk, pk
            )
            entity = self.get_object()
            logger.debug("Got entity: %s.", entity.pk)
            # Read and validate the request data.
            serializer = AnnotationsByPathSerializer(data=request.data, many=True)
            serializer.is_valid(raise_exception=True)
            annotations = {
                value["field_path"]: value["value"] for value in serializer.data
            }
            entity.update_annotations(annotations, request.user)
        except Exception:
            logger.exception("Set annotation exception")
            raise
        return Response()



""".. Ignore pydocstyle D400.

============
Flow Filters
============

"""

import re
import types
from copy import deepcopy
from dataclasses import dataclass
from functools import partial, reduce
from typing import Any, Optional, Union

from django.contrib.auth import get_user_model
from django.contrib.auth.models import Group
from django.contrib.postgres.search import SearchQuery, SearchRank
from django.core.exceptions import FieldDoesNotExist, ValidationError
from django.db import connection
from django.db.models import Count, F, ForeignKey, Model, OuterRef, Q, Subquery
from django.db.models.expressions import RawSQL
from django.db.models.query import QuerySet
from django_filters import rest_framework as filters
from django_filters.constants import EMPTY_VALUES
from django_filters.filterset import FilterSetMetaclass
from rest_framework import fields
from rest_framework.filters import OrderingFilter as DrfOrderingFilter

from resolwe.composer import composer
from resolwe.flow.models.fields import VersionField
from resolwe.permissions.models import (
    Permission,
    PermissionInterface,
    get_anonymous_user,
)

from .models import (
    AnnotationField,
    AnnotationPreset,
    AnnotationValue,
    Collection,
    Data,
    DescriptorSchema,
    Entity,
    Process,
    Relation,
)
from .models.annotations import AnnotationType

# The key used in the full-text search filter.
FULL_TEXT_SEARCH_KEY = "text"
# The actual database fields used for full-text search for every supported entity.
FULL_TEXT_SEARCH_FIELD = {"Entity": "search", "Data": "search", "Collection": "search"}

RELATED_LOOKUPS = [
    "exact",
    "in",
    "isnull",
]
NUMBER_LOOKUPS = [
    "exact",
    "in",
    "gt",
    "gte",
    "lt",
    "lte",
    "isnull",
]
TEXT_LOOKUPS = [
    "exact",
    "iexact",
    "contains",
    "icontains",
    "in",
    "startswith",
    "istartswith",
    "isnull",
]
DATE_LOOKUPS = [
    "exact",
    "gt",
    "gte",
    "lt",
    "lte",
    "isnull",
]
DATETIME_LOOKUPS = DATE_LOOKUPS + [
    "date",
    "time",
]
SLUG_LOOKUPS = [
    "exact",
    "in",
]


user_model = get_user_model()


class CheckQueryParamsMixin:
    """Custom query params validation."""

    def get_always_allowed_arguments(self):
        """Get always allowed query arguments."""
        return (
            "fields",
            "format",
            "limit",
            "offset",
            "ordering",
            # Used for full-text search.
            "text",
        )

    def validate_query_params(self):
        """Ensure no unsupported query params were used."""
        allowed_params = set(self.get_filters().keys())
        allowed_params.update(self.get_always_allowed_arguments())

        unallowed = set(self.request.query_params.keys()) - allowed_params

        if unallowed:
            msg = (
                "Unsupported parameter(s): {}. Please use a combination of: {}.".format(
                    ", ".join(unallowed), ", ".join(allowed_params)
                )
            )
            self.form.add_error(field=None, error=msg)

    def is_valid(self):
        """Validate filterset."""
        self.validate_query_params()
        return super().is_valid()


class UserFilterMixin:
    """Mixin for filtering by contributors and owners."""

    @staticmethod
    def _get_user_subquery(value: str):
        user_subquery = user_model.objects.all()
        for key in value.split():
            user_subquery = user_subquery.extra(
                where=["CONCAT_WS(' ', first_name, last_name, username) ILIKE %s"],
                params=("%{}%".format(key),),
            )

        return user_subquery

    def filter_owners(self, queryset: QuerySet, name: str, value: int):
        """Filter queryset by owner's id."""
        try:
            user = user_model.objects.get(pk=value)
        except user_model.DoesNotExist:
            return queryset.none()

        return queryset.filter_for_user(
            user, self.owner_permission, with_superuser=False
        )

    def filter_owners_name(self, queryset: QuerySet, name: str, value: str):
        """Filter queryset by owner's name."""
        result = queryset.model.objects.none()
        user_subquery = self._get_user_subquery(value)
        for user in user_subquery:
            result = result.union(
                queryset.filter_for_user(
                    user, self.owner_permission, with_superuser=False
                )
            )

        # Union can no longer be filtered, so we have to create a new queryset
        # for following filters.
        return result.model.objects.filter(pk__in=Subquery(result.values("pk")))

    def filter_contributor_name(self, queryset: QuerySet, name: str, value: str):
        """Filter queryset by owner's name."""
        return queryset.filter(contributor__in=self._get_user_subquery(value))

    def filter_for_user(
        self, queryset: QuerySet[PermissionInterface], name: str, value: str
    ) -> QuerySet[PermissionInterface]:
        """Filter queryset by permissions."""
        user = self.request.user
        return queryset.filter_for_user(user, Permission.from_name(value))

    def filter_for_group(self, queryset: QuerySet, name: str, value: Group):
        """Filter queryset by group."""
        return queryset.filter(permission_group__permissions__group=value)

    def filter_private(self, queryset: QuerySet, name: str, value: bool):
        """Return only elements that are not public (explicitly shared with me)."""

        public_filter = Q(permission_group__permissions__user=get_anonymous_user())

        # Exclude public elements if filter value is True
        if value:
            return queryset.exclude(public_filter)
        else:
            return queryset.filter(public_filter)


class TagsFilter(filters.BaseCSVFilter, filters.CharFilter):
    """Filter for tags."""

    def __init__(self, *args, **kwargs):
        """Construct tags filter."""
        kwargs.setdefault("lookup_expr", "contains")
        super().__init__(*args, **kwargs)


class ResolweFilterMetaclass(FilterSetMetaclass):
    """Support extending filter by third-party packages.

    .. note::

        This injects filters in an undocumented way and can thus break in
        future releases.

    For example, to add a filter by ``my_field``, define an extension and add
    it to the composer:

    .. code-block:: python

        from resolwe.composer import composer


        class ExtendedFilter:

            my_field = filters.CharFilter()

        composer.add_extension("path.to.my.Filter", ExtendedFilter)

    """

    def __new__(
        mcs: type[object],
        name: str,
        bases: tuple[type[object]],
        namespace: dict[str, Any],
    ):
        """Inject filters."""

        def inject_composer_extensions(namespace: dict[str, Any], name: str):
            """Inject filters from composer extensions."""
            class_path = "{}.{}".format(namespace["__module__"], name)
            for extension in composer.get_extensions(class_path):
                for name in dir(extension):
                    if name.startswith("__"):
                        continue
                    namespace[name] = getattr(extension, name)

        def get_related_model_from_path(related_path: str) -> type[Model]:
            """Get the related model from the path."""
            return reduce(
                lambda Model, part: Model._meta.get_field(part).related_model,
                related_path.split("__"),
                namespace["Meta"].model,
            )

        def add_filters_with_permissions(
            prefix: str, related_path: str, base_filter: filters.FilterSet
        ):
            """Add filters on related objects with permissions.

            All the filters belonging to the BaseFilter are added with the given prefix.
            """

            def filter_permissions(
                self,
                qs: QuerySet,
                value: str,
                original_filter: filters.Filter,
                original_model: Model,
            ):
                """Apply the filter and respect permissions."""

                # Do not filter when value is empty. At least one of the values must be
                # non-empty since form in the AnnotationValueFilter class requires it.
                if value in EMPTY_VALUES:
                    return qs

                visible_objects = list(
                    original_filter.filter(original_model.objects.all(), value)
                    .filter_for_user(self.parent.request.user)
                    .values_list("pk", flat=True)
                )
                return qs.filter(**{f"{related_path}__in": visible_objects})

            # Add all filters from EntityFilter to namespaces before creating class.
            BaseModel = get_related_model_from_path(related_path)
            for filter_name, filter in base_filter.get_filters().items():
                new_filter_name = f"{prefix}__{filter_name}"
                if filter_name == "id" or filter_name.startswith("id__"):
                    new_filter_name = f"{prefix}{filter_name[2:]}"

                filter_copy = deepcopy(filter)
                filter_copy.field_name = f"{prefix}__{filter_copy.field_name}"
                filter_method = partial(
                    filter_permissions, original_filter=filter, original_model=BaseModel
                )
                # Bind the new_filter to filter instance and set it as new filter.
                filter_copy.filter = types.MethodType(filter_method, filter_copy)
                namespace[new_filter_name] = filter_copy
                # If filter uses a method, add it to the namespace as well.
                if filter_copy.method is not None:
                    namespace[filter_copy.method] = deepcopy(
                        getattr(base_filter, filter_copy.method)
                    )

        def inject_related_permissions(namespace: dict):
            """Parse the namespace and add filters with permissions."""
            for prefix, value in list(namespace.items()):
                if isinstance(value, FilterRelatedWithPermissions):
                    namespace.pop(prefix)
                    add_filters_with_permissions(
                        prefix, value.related_path or prefix, value.BaseFilter
                    )

        inject_related_permissions(namespace)
        inject_composer_extensions(namespace, name)
        return super().__new__(mcs, name, bases, namespace)


@dataclass
class FilterRelatedWithPermissions:
    """Base class for filters with permissions.

    The class using this must have FilterRelatedWithPermissionsMeta as its meta class.
    The filter is then defined as:

    prefix = FilterRelatedWithPermissions(BaseFilter, related_path="related__path")

    All the filters from the BaseFilter (with the given prefix) are added and
    permissions on the related objects are respected.

    The optional parameter `related__path` is used to specify the path of the related
    object (if different from the prefix).
    """

    BaseFilter: type[filters.FilterSet]
    related_path: Optional[str] = None


class BaseResolweFilter(
    CheckQueryParamsMixin, filters.FilterSet, metaclass=ResolweFilterMetaclass
):
    """Base filter for Resolwe's endpoints."""

    owner_permission = Permission.OWNER

    class Meta:
        """Filter configuration."""

        fields = {
            "contributor": ["exact", "in"],
            "created": DATETIME_LOOKUPS[:],
            "id": NUMBER_LOOKUPS[:],
            "name": TEXT_LOOKUPS[:],
            "modified": DATETIME_LOOKUPS[:],
            "slug": SLUG_LOOKUPS[:],
            "version": NUMBER_LOOKUPS[:],
        }

        filter_overrides = {
            VersionField: {"filter_class": filters.CharFilter},
            ForeignKey: {"filter_class": filters.Filter},
        }


class DescriptorSchemaFilter(BaseResolweFilter):
    """Filter the DescriptorSchema endpoint."""

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = DescriptorSchema


class CharInFilter(filters.BaseInFilter, filters.CharFilter):
    """Basic filter for CharField with 'in' lookup."""


class BaseCollectionFilter(UserFilterMixin, BaseResolweFilter):
    """Base filter for Collection and Entity endpoints."""

    contributor_name = filters.CharFilter(method="filter_contributor_name")
    owners = filters.CharFilter(method="filter_owners")
    owners_name = filters.CharFilter(method="filter_owners_name")
    shared_with_me = filters.BooleanFilter(method="filter_private")
    group = filters.ModelChoiceFilter(
        method="filter_for_group", queryset=Group.objects.all()
    )
    permission = filters.CharFilter(method="filter_for_user")
    tags = TagsFilter()
    status = filters.CharFilter(field_name="status")
    status__in = CharInFilter(field_name="status", lookup_expr="in")

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        fields = {
            **BaseResolweFilter.Meta.fields,
            **{
                "description": TEXT_LOOKUPS[:],
            },
        }


class CollectionFilter(BaseCollectionFilter):
    """Filter the Collection endpoint."""

    entity_count = filters.NumberFilter(
        method="count_entities", field_name="entity__count"
    )
    entity_count__gt = filters.NumberFilter(
        method="count_entities", field_name="entity__count__gt"
    )
    entity_count__lt = filters.NumberFilter(
        method="count_entities", field_name="entity__count__lt"
    )
    entity_count__gte = filters.NumberFilter(
        method="count_entities", field_name="entity__count__gte"
    )
    entity_count__lte = filters.NumberFilter(
        method="count_entities", field_name="entity__count__lte"
    )

    class Meta(BaseCollectionFilter.Meta):
        """Filter configuration."""

        model = Collection
        fields = {
            **BaseCollectionFilter.Meta.fields,
            **{
                "descriptor_schema": ["exact"],
            },
        }

    def count_entities(self, queryset: QuerySet, name: str, value: str):
        """Filter by the number of associated entities."""

        return queryset.annotate(Count("entity")).filter(**{name: value})


class EntityFilter(BaseCollectionFilter):
    """Filter the Entity endpoint."""

    relation_id = filters.NumberFilter(field_name="relation__id")
    annotations = filters.CharFilter(method="filter_annotations")

    def filter_annotations(self, queryset: QuerySet, name: str, value: str):
        """Filter entities by annotations.

        The value must be in the following format: field_id__lookup_type:value
        """
        type_to_field = {
            AnnotationType.DATE.value: fields.DateField(),
            AnnotationType.STRING.value: fields.CharField(),
            AnnotationType.INTEGER.value: fields.IntegerField(),
            AnnotationType.DECIMAL.value: fields.FloatField(),
        }
        allowed_lookups = {
            AnnotationType.DATE.value: DATE_LOOKUPS,
            AnnotationType.STRING.value: TEXT_LOOKUPS,
            AnnotationType.INTEGER.value: NUMBER_LOOKUPS,
            AnnotationType.DECIMAL.value: NUMBER_LOOKUPS,
        }
        post_processing_map = {
            AnnotationType.DATE.value: lambda value: value.isoformat()
        }

        match = re.match(
            r"^(?P<field_id>\d+)(__(?P<lookup>\w+))?:(?P<value>.+)$", value
        )
        if not match:
            raise ValidationError(f"Invalid annotation filter format '{value}'.")

        field_id = match.group("field_id")
        lookup_type = match.group("lookup") or "exact"
        string_value = match.group("value")
        field_type = (
            AnnotationField.objects.filter(id=field_id)
            .values_list("type", flat=True)
            .get()
        )
        field = type_to_field[field_type]

        if lookup_type not in allowed_lookups[field_type]:
            raise ValueError(
                f"Lookup '{lookup_type}' not supported for the field '{field_id}'."
            )

        processing = field.run_validation
        if post_processing := post_processing_map.get(field_type, None):
            processing = lambda value: post_processing(field.run_validation(value))

        if lookup_type == "in":
            validated_value = map(processing, string_value.split(","))
        else:
            validated_value = processing(string_value)

        q_filter = {
            f"annotations___value__value__{lookup_type}": validated_value,
            f"annotations___value__label__{lookup_type}": validated_value,
        }
        value_query = Q()
        for key, value in q_filter.items():
            value_query |= Q(**{key: value})
        field_query = Q(annotations__field_id=field_id)
        return queryset.filter().filter(field_query & value_query)

    class Meta(BaseCollectionFilter.Meta):
        """Filter configuration."""

        model = Entity
        fields = {
            **BaseCollectionFilter.Meta.fields,
            **{
                "collection": RELATED_LOOKUPS[:],
                "collection__name": TEXT_LOOKUPS[:],
                "collection__slug": SLUG_LOOKUPS[:],
                "type": ["exact"],
            },
        }


class ProcessFilter(BaseResolweFilter):
    """Filter the Process endpoint."""

    category = filters.CharFilter(field_name="category", lookup_expr="startswith")
    type = filters.CharFilter(field_name="type", lookup_expr="startswith")

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Process
        fields = {
            **BaseResolweFilter.Meta.fields,
            **{
                "is_active": ["exact"],
                "scheduling_class": ["exact"],
            },
        }


class DataFilter(UserFilterMixin, BaseResolweFilter):
    """Filter the Data endpoint."""

    contributor_name = filters.CharFilter(method="filter_contributor_name")
    owners = filters.CharFilter(method="filter_owners")
    owners_name = filters.CharFilter(method="filter_owners_name")
    tags = TagsFilter()
    type = filters.CharFilter(field_name="process__type", lookup_expr="startswith")
    type__exact = filters.CharFilter(field_name="process__type", lookup_expr="exact")
    relation_id = filters.NumberFilter(
        field_name="entity__relationpartition__relation_id"
    )
    shared_with_me = filters.BooleanFilter(method="filter_private")
    group = filters.ModelChoiceFilter(
        method="filter_for_group", queryset=Group.objects.all()
    )
    permission = filters.CharFilter(method="filter_for_user")

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Data
        fields = {
            **BaseResolweFilter.Meta.fields,
            **{
                "collection": RELATED_LOOKUPS[:],
                "collection__name": TEXT_LOOKUPS[:],
                "collection__slug": SLUG_LOOKUPS[:],
                "entity": RELATED_LOOKUPS[:],
                "entity__name": TEXT_LOOKUPS[:],
                "entity__slug": SLUG_LOOKUPS[:],
                "process": RELATED_LOOKUPS[:],
                "process__name": TEXT_LOOKUPS[:],
                "process__slug": SLUG_LOOKUPS[:],
                "finished": DATETIME_LOOKUPS[:],
                "started": DATETIME_LOOKUPS[:],
                "status": ["exact", "in"],
            },
        }


class RelationFilter(BaseResolweFilter):
    """Filter the Relation endpoint."""

    category = filters.CharFilter(lookup_expr="iexact")
    collection = filters.ModelChoiceFilter(queryset=Collection.objects.all())
    type = filters.CharFilter(field_name="type__name")
    tags = TagsFilter(field_name="collection__tags")

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = Relation
        fields = {
            **BaseResolweFilter.Meta.fields,
            **{
                "descriptor_schema": ["exact"],
            },
        }

    def get_always_allowed_arguments(self):
        """Get always allowed query arguments."""
        return super().get_always_allowed_arguments() + (
            "entity",
            "label",
            "position",
        )


class OrderingFilter(DrfOrderingFilter):
    """Order results by field specified in request.

    If no ordering is given, default ordering is used. Default ordering
    is skipped for views filtered by full-text search as it already
    applies ordering by ranking.
    """

    @property
    def _annotation_handlers(self):
        """Return a list of custom annotation handlers."""
        return (self._annotate_annotations,)

    def _annotate_annotations(self, queryset, ordering):
        """Annotate queryset with annotations.

        The annotations are only applied if ordering by annotations is requested.
        """
        if not issubclass(queryset.model, Entity):
            return queryset

        regex = re.compile(r"-?annotations_order_(?P<field_id>\d+)")
        for field in ordering:
            if not isinstance(field, str):
                continue
            if match := regex.match(field):
                field_id = match.group("field_id")
                # Always filter by user-visible attribute annotation_label.
                annotation_value = AnnotationValue.objects.filter(
                    entity_id=OuterRef("pk"), field_id=field_id
                ).values("_value__label")
                annotate = {f"annotations_order_{field_id}": Subquery(annotation_value)}
                queryset = queryset.annotate(**annotate)
        return queryset

    def _annotate_queryset(self, queryset, ordering):
        """Annotate queryset with ordering fields.

        Some ordering field require additional annotations to be present.
        """
        for handler in self._annotation_handlers:
            queryset = handler(queryset, ordering)
        return queryset

    def filter_queryset(self, request, queryset, view):
        """Return ordered queryset."""
        if ordering := self.get_ordering(request, queryset, view):
            queryset = self._annotate_queryset(queryset, ordering).order_by(*ordering)
        return queryset

    def _remove_invalid_annotation_field(self, queryset, field, view, request):
        """Check if the given sorting field is annotations field.

        The format of the ordering field is annotations__field_id, possibly prefixed by
        a minus sign. The minus sign indicates descending order.
        """
        if re.match(r"-?annotations__(?P<field_id>\d+)", field):
            return field.replace("__", "_order_")

    def _remove_invalid_json_field(self, queryset, field, view, request):
        """Remove invalid JSON field.

        Either return None or a ordering expression.
        """
        # When a dot is present in the order field, treat it as a JSON path.
        if "." not in field:
            return None

        path = field.split(".")
        base_field = path[0]
        if len(path) < 2:
            return None
        if not super().remove_invalid_fields(queryset, [base_field], view, request):
            return None

        reverse = False
        if base_field[0] == "-":
            base_field = base_field[1:]
            reverse = True

        # Resolve base field.
        try:
            quote_name = connection.ops.quote_name
            model_meta = queryset.model._meta
            base_field = "{}.{}".format(
                quote_name(model_meta.db_table),
                quote_name(model_meta.get_field(base_field).column),
            )
        except FieldDoesNotExist:
            return None

        placeholders = "->".join(["%s"] * (len(path) - 2))
        # The last placeholder should be accessed via the ->> operator.
        if placeholders:
            placeholders = "->{}->>%s".format(placeholders)
        else:
            placeholders = "->>%s"

        expression = RawSQL(
            # We can use base_field here directly because we've resolved it via Django ORM.
            "{}{}".format(base_field, placeholders),
            params=path[1:],
        )
        if reverse:
            expression = expression.desc()

        return expression

    @property
    def _invalid_field_handlers(self):
        """Return a list of custom invalid field handlers."""
        return (
            self._remove_invalid_annotation_field,
            self._remove_invalid_json_field,
        )

    def _remove_invalid_field(self, queryset, field, view, request):
        """Remove invalid fields from the ordering.

        Returns the ordering expression if the field is valid, otherwise None.
        """
        arguments = (queryset, field, view, request)
        for handler in self._invalid_field_handlers:
            if field := handler(*arguments):
                return field
        return None

    def remove_invalid_fields(self, queryset, fields, view, request):
        """Remove invalid fields and preserve sort order.

        First check for custom ordering fields and then for the default ones.
        """
        order_fields = []
        base_order_fields = super().remove_invalid_fields(
            queryset, fields, view, request
        )
        for field in fields:
            if field in base_order_fields:
                order_fields.append(field)
            elif expression := self._remove_invalid_field(
                queryset, field, view, request
            ):
                order_fields.append(expression)
            else:
                raise ValidationError(
                    "Ordering by field '{}' is not supported.".format(field)
                )
        return order_fields


class FullTextSearchFilter(DrfOrderingFilter):
    """Full text search filter.

    Filter and order the queryset by the full text search rank.
    """

    def filter_queryset(self, request, queryset, view):
        """Return ordered queryset.

        If full text search is used and ordering by some other field is requested, raise
        an error.
        """
        if FULL_TEXT_SEARCH_KEY in view.request.query_params:

            order_fields = request.query_params.get(self.ordering_param)
            if FULL_TEXT_SEARCH_KEY in view.request.query_params and order_fields:
                raise ValidationError(
                    f"Ordering by full text search rank and other fields ({order_fields}) "
                    f"is not supported."
                )

            model_name = queryset.model.__name__
            field_name = FULL_TEXT_SEARCH_FIELD.get(model_name)
            if field_name is None:
                raise ValidationError(
                    f"Full text search is not supported for the model {model_name}."
                )
            value = view.request.query_params[FULL_TEXT_SEARCH_KEY]
            query = SearchQuery(value, config="simple_unaccent")
            return (
                queryset.filter(**{field_name: query})
                .annotate(rank=SearchRank(F(field_name), query))
                .order_by("-rank")
            )
        return queryset


class AnnotationFieldFilter(BaseResolweFilter):
    """Filter the AnnotationField endpoint."""

    entity = FilterRelatedWithPermissions(EntityFilter, related_path="values__entity")

    @classmethod
    def filter_for_field(cls, field, field_name, lookup_expr=None):
        """Add permission check for collections lookups.

        The old filter method is moved to the original_filter and replaced with the
        filter that checks the permissions on the collections.
        """

        def collection_permission_filter(
            self, qs: QuerySet, value: Union[str, list[Collection]]
        ):
            """Check if user has access the collections.

            The collection filer is defined in the parent filter.

            :raises PermissionDenied: if user does not have the permissions to access any
                of the collections defined by the filter.
            """
            qs = self._original_filter(qs, value)
            if not value:
                return qs

            field_name = self.field_name
            exact_filter = field_name == "collection"
            if exact_filter:
                field_name += "__id"

            collection_field = field_name.split("__", maxsplit=1)[1]
            collection_filter = f"{collection_field}__{self.lookup_expr}"
            # Exact filter is special: value is a list of collections.
            filter_value = value[0].pk if exact_filter else value
            visible_collections = Collection.objects.filter(
                **{collection_filter: filter_value}
            ).filter_for_user(self.parent.request.user)
            qs = qs.filter(collection__in=visible_collections)
            return qs

        base_filter = filters.FilterSet.filter_for_field(field, field_name, lookup_expr)
        # When collection filter is used replace it with the one that respects
        # permissions on collections.
        if field_name.startswith("collection"):
            base_filter._original_filter = base_filter.filter
            base_filter.filter = partial(collection_permission_filter, base_filter)
        return base_filter

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = AnnotationField
        fields = {
            **{
                "id": NUMBER_LOOKUPS[:],
                "name": TEXT_LOOKUPS[:],
                "label": TEXT_LOOKUPS[:],
                "type": TEXT_LOOKUPS[:],
                "description": TEXT_LOOKUPS[:],
                "group__name": TEXT_LOOKUPS[:],
                "collection": ["exact"],
                "collection__name": TEXT_LOOKUPS[:],
                "collection__slug": SLUG_LOOKUPS[:],
                "required": ["exact"],
            },
        }


class AnnotationPresetFilter(BaseResolweFilter):
    """Filter the AnnotationPreset endpoint."""

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = AnnotationPreset
        fields = {
            **{
                "name": TEXT_LOOKUPS,
            },
        }


class AnnotationValueFilter(BaseResolweFilter):
    """Filter the AnnotationValue endpoint."""

    entity = FilterRelatedWithPermissions(EntityFilter)
    label = filters.CharFilter(method="filter_by_label")

    def filter_by_label(self, queryset: QuerySet, name: str, value: str):
        """Filter by label."""
        return queryset.filter(_value__label__icontains=value)

    class Meta(BaseResolweFilter.Meta):
        """Filter configuration."""

        model = AnnotationValue
        fields = {
            **{
                "id": NUMBER_LOOKUPS[:],
                "field__id": NUMBER_LOOKUPS[:],
                "field__name": TEXT_LOOKUPS[:],
                "field__label": TEXT_LOOKUPS[:],
                "field__group__name": TEXT_LOOKUPS[:],
                "created": DATE_LOOKUPS[:],
                "contributor": ["exact", "in"],
            },
        }



"""Collection viewset."""

from typing import Optional

from django.db import transaction
from django.db.models import F, Func, OuterRef, Prefetch, QuerySet, Subquery
from django.db.models.functions import Coalesce
from drf_spectacular.utils import extend_schema
from rest_framework import exceptions, mixins, request, serializers, status, viewsets
from rest_framework.decorators import action
from rest_framework.permissions import SAFE_METHODS
from rest_framework.response import Response

from resolwe.flow.filters import CollectionFilter
from resolwe.flow.models import Collection, Data, DescriptorSchema, Entity
from resolwe.flow.serializers import CollectionSerializer
from resolwe.flow.serializers.annotations import AnnotationFieldDictSerializer
from resolwe.observers.mixins import ObservableMixin
from resolwe.permissions.loader import get_permissions_class
from resolwe.permissions.mixins import ResolwePermissionsMixin
from resolwe.permissions.models import PermissionModel

from .mixins import (
    ResolweBackgroundDeleteMixin,
    ResolweBackgroundDuplicateMixin,
    ResolweCheckSlugMixin,
    ResolweCreateModelMixin,
    ResolweUpdateModelMixin,
)


class BaseCollectionViewSet(
    ResolweCreateModelMixin,
    mixins.RetrieveModelMixin,
    ResolweUpdateModelMixin,
    ResolweBackgroundDeleteMixin,
    mixins.ListModelMixin,
    ResolwePermissionsMixin,
    ResolweCheckSlugMixin,
    ResolweBackgroundDuplicateMixin,
    viewsets.GenericViewSet,
):
    """Base API view for :class:`Collection` objects."""

    qs_permission_model = PermissionModel.objects.select_related("user", "group")

    # Add entity count as a subquery: by default Django joins tables and this explodes
    # when one tries to annotate the collection query with data and entity count at the
    # same time.
    entity_count_subquery = (
        Entity.objects.filter(collection_id=OuterRef("pk"))
        .annotate(count=Func(F("id"), function="Count"))
        .values("count")
    )
    data_count_subquery = (
        Data.objects.filter(collection_id=OuterRef("pk"))
        .annotate(count=Func(F("id"), function="Count"))
        .values("count")
    )
    data_status_subquery = (
        Data.objects.filter(collection_id=OuterRef("pk"))
        .annotate(
            count=Coalesce(
                Func(
                    F("status"),
                    function="ARRAY_AGG",
                    template="%(function)s(DISTINCT %(expressions)s)",
                ),
                [],
            )
        )
        .values("count")
    )

    qs_descriptor_schema = DescriptorSchema.all_objects.select_related("contributor")

    queryset = (
        Collection.objects.select_related("contributor")
        .prefetch_related(Prefetch("descriptor_schema", queryset=qs_descriptor_schema))
        .annotate(data_statuses=Subquery(data_status_subquery))
        .annotate(data_count=Subquery(data_count_subquery))
        .annotate(entity_count=Subquery(entity_count_subquery))
    )

    filterset_class = CollectionFilter
    permission_classes = (get_permissions_class(),)

    ordering_fields = (
        "contributor",
        "contributor__first_name",
        "contributor__last_name",
        "collection__name",
        "created",
        "id",
        "modified",
        "name",
        "status",
    )
    ordering = "id"

    def get_queryset(self) -> QuerySet:
        """Prefetch permissions for current user."""
        # Only annotate the queryset with status on safe methods. When updating
        # the annotation interfers with update (as it adds group by statement).
        if self.request.method in SAFE_METHODS:
            self.queryset = self.queryset.annotate_status()
        return self.prefetch_current_user_permissions(self.queryset)

    def create(self, request: request.Request, *args, **kwargs) -> Response:
        """Only authenticated users can create new collections."""
        if not request.user.is_authenticated:
            raise exceptions.NotFound

        return super().create(request, *args, **kwargs)


class CollectionViewSet(ObservableMixin, BaseCollectionViewSet):
    """API view for :class:`Collection` objects."""

    serializer_class = CollectionSerializer

    def _set_fields(self, request: request.Request, property_name: str) -> Response:
        """Set the fields on the collection."""
        collection = self.get_object()
        # Validate that the collection has the specified property with a callable 'set' method.
        if not hasattr(collection, property_name):
            raise serializers.ValidationError(
                {
                    property_name: f"The collection does not have a '{property_name}' attribute."
                }
            )
        attribute = getattr(collection, property_name)
        if not callable(getattr(attribute, "set", None)):
            raise serializers.ValidationError(
                {
                    property_name: f"The '{property_name}' attribute does not have a callable 'set' method."
                }
            )
        # Read and validate the request data. Get the serializer from the DRF viewset
        # since the get_serializer method is overridden in the collection viewset class.
        SerializerClass = viewsets.GenericViewSet.get_serializer_class(self)
        serializer = SerializerClass(data=request.data)
        # Validate the serializer data.
        serializer.is_valid(raise_exception=True)
        getattr(collection, property_name).set(serializer.validated_data[property_name])
        return Response(status=status.HTTP_204_NO_CONTENT)

    @extend_schema(
        request=AnnotationFieldDictSerializer(),
        responses={status.HTTP_204_NO_CONTENT: None},
    )
    @action(
        detail=True, methods=["post"], serializer_class=AnnotationFieldDictSerializer
    )
    @transaction.atomic
    def set_annotation_fields(
        self, request: request.Request, pk: Optional[int] = None
    ) -> Response:
        """Set AnnotationFields on collection."""
        return self._set_fields(request, "annotation_fields")
